#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
from typing import Callable, List, Optional, Sequence, Tuple, Union

from highdicom.content import (IssuerOfIdentifier, SpecimenCollection,
                               SpecimenDescription, SpecimenPreparationStep,
                               SpecimenSampling, SpecimenStaining)
from opentile.common import Tiler
from opentile.interface import OpenTile
from pydicom.dataset import Dataset
from pydicom.sequence import Sequence as DicomSequence
from pydicom.sr.coding import Code
from pydicom.uid import UID as Uid
from pydicom.uid import generate_uid
from wsidicom.conceptcode import (AnatomicPathologySpecimenTypesCode,
                                  ConceptCode, SpecimenEmbeddingMediaCode,
                                  SpecimenFixativesCode,
                                  SpecimenPreparationProcedureCode,
                                  SpecimenSamplingProcedureCode,
                                  SpecimenStainsCode)


def get_image_type(image_flavor: str, level_index: int) -> List[str]:
    """Return image type.

    Parameters
    ----------
    image_flavor: str
        Image flavor ('VOLUME', 'LABEL', 'OVERVIEW')
    level_index: int:
        Pyramidal level index of the image.

    Returns
    ----------
    List[str]
        Image type.
    """
    if image_flavor == 'VOLUME' and level_index == 0:
        resampled = 'NONE'
    else:
        resampled = 'RESAMPLED'

    return ['ORGINAL', 'PRIMARY', image_flavor, resampled]


def create_base_dataset(
    modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
) -> Dataset:
    """Create a base dataset by combining module datasets with a minimal
    wsi dataset.

    Parameters
    ----------
    modules: Optional[Union[Dataset, Sequence[Dataset]]] = None

    Returns
    ----------
    Dataset
        Combined base dataset.
    """
    base_dataset = create_wsi_dataset()
    if modules is None:
        modules = create_default_modules()
    if isinstance(modules, Sequence):
        for module in modules:
            base_dataset.update(module)
    elif isinstance(modules, Dataset):
        base_dataset.update(modules)
    else:
        raise TypeError(
            'datasets parameter should be single or list of Datasets'
        )
    return base_dataset


def populate_base_dataset(
    tiler: Tiler,
    base_dataset: Dataset,
    include_confidential: bool = True
) -> Dataset:
    """Populate dataset with properties from tiler, if present.
    Parameters
    ----------
    tiler: Tiler
        A opentile Tiler.
    base_dataset: Dataset
        Dataset to append properties to.
    include_confidential: bool = True
        If to include confidential properties (see https://dicom.nema.org/medical/dicom/current/output/html/part15.html#table_E.1-1)  # NOQA

    Returns
    ----------
    Dataset
        Dataset with added properties.
    """
    for property, value in tiler.properties.items():
        if property == 'aquisition_datetime' and include_confidential:
            base_dataset.AcquisitionDateTime = value
        elif property == 'device_serial_number' and include_confidential:
            base_dataset.DeviceSerialNumber = value
        elif property == 'manufacturer':
            base_dataset.Manufacturer = value
        elif property == 'model':
            base_dataset.ManufacturerModelName = value
        elif property == 'software_versions':
            base_dataset.SoftwareVersions = value
        elif property == 'lossy_image_compression_method':
            base_dataset.LossyImageCompressionMethod = value
        elif property == 'lossy_image_compression_ratio':
            base_dataset.LossyImageCompressionRatio = value
        elif property == 'photometric_interpretation':
            base_dataset.PhotometricInterpretation = value
    return base_dataset


def create_wsi_dataset(
    offset: Tuple[float, float] = (0.0, 0.0),
    uid_generator: Callable[..., Uid] = generate_uid
) -> Dataset:
    """Return dataset containing (parts of) SOP common, general series, whole
    slide icrosocy series, frame of reference, acquisition context, multi-frame
    dimension, and whole slide microscopy image modules.

    Some modules returned 'not complete', and completed during image data
    import or file save():
        SOP common module:
            SOPInstanceUID
        Whole slide microscopy image module:
            ImageType
            Aquisition DateTime
            SamplesPerPixel
            PhotometricInterpretation
            PlanarConfiguration
            NumberOfFrames
            BitsAllocated
            BitsStored
            HighBit
            PixelRepresentation
            LossyImageCompression (and conditionals)
            ImagedVolumeWidth
            ImagedVolumeWidth
            ImagedVolumeDepth
            TotalPixelMatrixColumns
            TotalPixelMatrixRows
            FocusMethod
            ExtendedDepthOfField (and conditionals)

    Parameters
    ----------
    image_offset: Tuple[float, float] = 0.0
        X and Y offset (in mm) to first pixel in image data.
    uid_generator: Callable[..., Uid] = generate_uid
        Function that can generate Uids.

    Returns
    ----------
    Dataset
        WSI dataset.
    """
    dataset = Dataset()

    # SOP common module (SOPInstanceUID written on save())
    dataset.SOPClassUID = '1.2.840.10008.5.1.4.1.1.77.1.6'

    # dataset.StudyInstanceUID = uid_generator()
    # General series and Whole slide Microscopy modules
    dataset.SeriesInstanceUID = uid_generator()
    dataset.Modality = 'SM'

    # Frame of reference module
    dataset.FrameOfReferenceUID = uid_generator()

    # Acquisition context module (empty)
    dataset.AcquisitionContextSequence = DicomSequence()

    # Multi-frame Dimension module
    dimension_organization_uid = uid_generator()
    dimension_organization_sequence = Dataset()
    dimension_organization_sequence.DimensionOrganizationUID = (
        dimension_organization_uid
    )
    dataset.DimensionOrganizationSequence = DicomSequence(
        [dimension_organization_sequence]
    )

    # Whole slide micropscopy image module (most filled when importing file)
    dataset.BurnedInAnnotation = 'NO'
    dataset.SpecimenLabelInImage = 'NO'
    dataset.VolumetricProperties = 'VOLUME'
    dataset.ImageOrientationSlide = [-1, 0, 0, 0, -1, 0]
    offset_item = Dataset()
    offset_item.x_offset = offset[0]
    offset_item.y_offset = offset[1]
    dataset.TotalPixelMatrixOriginSequence = DicomSequence([offset_item])
    return dataset


def create_study_module(
    study_id: str = "",
    study_date: Optional[datetime.date] = None,
    study_time: Optional[datetime.time] = None,
    study_accession_number: str = "",
    referring_physician_name: str = "",
    uid_generator: Callable[..., Uid] = generate_uid
) -> Dataset:
    """Create general study module.

    Parameters
    ----------
    study_id: str = ""
        Study identifier. Can be empty.
    Optional[datetime.date] = None
        Date the study started. Can be empty.
    Optional[datetime.time] = None
        Time the study started. Can be empty.
    accession_number: str = ""
        Order for the study. Can be empty.
    referring_physician_name: str = ""
        Name of reffering physician. Can be empty.

    Returns
    ----------
    Dataset
        Dataset containing general study module.
    """
    dataset = Dataset()
    dataset.StudyInstanceUID = uid_generator()
    dataset.StudyID = study_id
    dataset.StudyDate = study_date
    dataset.StudyTime = study_time
    dataset.AccessionNumber = study_accession_number
    dataset.ReferringPhysicianName = referring_physician_name
    return dataset


def create_patient_module(
    name: str = "",
    id: str = "",
    birth_date: Optional[datetime.date] = None,
    sex: str = "",
    identity_removed: bool = False,
    de_indentification_methods: Sequence[Union[str, Code]] = [],
    age_at_extraction: Optional[str] = None
) -> Dataset:
    """Create patient and patient study modules.

    Parameters
    ----------
    name: str = ""
        Patient's full name. Can be empty.
    id: str = ""
        Primary identifier for the Patient. Can be empty.
    birth_date: str = ""
        Birth date of the Patient. Can be empty.
    sex: str = ""
        Sex of the named Patient. F/M/O. Can be empty.
    identity_removed: bool = False
        True if partient idendity has been removed from attributes in dataset
        and in image data.
    de_indentification_methods: Sequence[Union[str, Code]] = []
        Method of de-indentification. Can be descriptive strings and/or coded
        values.
    age_at_extraction: Optional[str] = None
        Age of the Patient at extraction.

    Returns
    ----------
    Dataset
        Dataset containing patient and patient study modules.
    """
    dataset = Dataset()
    dataset.PatientName = name
    dataset.PatientID = id
    dataset.PatientBirthDate = birth_date
    dataset.PatientSex = sex
    dataset.PatientIdentityRemoved = 'YES' if identity_removed else 'NO'
    if identity_removed and de_indentification_methods is None:
        raise ValueError(
            'de-intification method must be specified if '
            'patient identity has been removed'
            )
    de_indentification_method_strings: List[str] = []
    de_indentification_method_codes: List[Code] = []

    for de_indentification_method in de_indentification_methods:
        if isinstance(de_indentification_method, str):
            de_indentification_method_strings.append(de_indentification_method)
        elif isinstance(de_indentification_method, Code):
            de_indentification_method_codes.append(de_indentification_method)
        else:
            raise TypeError(
                'De-indentification methods should be string or coded value'
            )
    if de_indentification_method_strings != []:
        dataset.DeidentificationMethod = de_indentification_method_strings
    if de_indentification_method_codes != []:
        dataset.DeidentificationMethodCodeSequence = DicomSequence(
            [
                ConceptCode.from_code(de_indentification_method_code).to_ds()
                for de_indentification_method_code
                in de_indentification_method_codes
            ]
        )
    if age_at_extraction is not None:
        dataset.PatientAge = age_at_extraction

    return dataset


def create_device_module(
    manufacturer: str = 'Unkown',
    model_name: str = 'Unkown',
    serial_number: str = 'Unkown',
    software_versions: Sequence[str] = ['Unkown']
) -> Dataset:
    """Create extended equipment module.

    Parameters
    ----------
    manufacturer: str = 'Unkown'.
        Manufacturer of the equipment.
    model_name: str = 'Unkown'
        Manufacturer's model name of the equipment.
    serial_number: str = 'Unkown'
        Manufacturer's serial number of the equipment.
    software_versions: Sequence[str] = ['Unkown']
        Software version of the equipment.

    Returns
    ----------
    Dataset
        Dataset containing extended equipment module.
    """
    dataset = Dataset()
    properties = {
        'Manufacturer': manufacturer,
        'ManufacturerModelName': model_name,
        'DeviceSerialNumber': serial_number,
        'SoftwareVersions': software_versions
    }
    for name, value in properties.items():
        setattr(dataset, name, value)
    return dataset


def create_sample(
    sample_id: str,
    stainings: Sequence[str],
    embedding_medium: Optional[str] = None,
    fixative: Optional[str] = None,
    specimen_id: Optional[str] = None,
    specimen_type: Optional[str] = None,
    specimen_sampling_method: Optional[str] = None,
    anatomical_sites: Optional[Sequence[Tuple[Code, Sequence[Code]]]] = None,
    location: Optional[Union[str, Tuple[float, float, float]]] = None,
    uid_generator: Callable[..., Uid] = generate_uid
) -> Dataset:
    """Create sample description.

    Parameters
    ----------
    sample_id: str
        Identifier for the sample.
    stainings: Sequence[str]
        Stainings used. See SpecimenStainsCode.list() for allowed values.
    embedding_medium: Optional[str] = None
        Embedding medium used. See SpecimenEmbeddingMediaCode.list() for
        allowed values.
    fixative: Optional[str] = None
        Fixative used. See SpecimenFixativesCode.list() for allowed values.
    specimen_id: Optional[str] = None
        Identifier for the specimen the sample was sampled from.
    specimen_type: Optional[str] = None
        Anatotomic type of specimen the sample was sampled from. See
        AnatomicPathologySpecimenTypesCode.list() for allowed values.
    specimen_sampling_method: Optional[str] = None
        Sampling method used for sampling the sample from the specimen. See
        SpecimenSamplingProcedureCode.list() for allowed values.
    anatomical_sites: Optional[Sequence[Tuple[Code, Sequence[Code]]]] = None
        List of original anatomical sites, each defined by a code and optional
        list of modifier codes.
    location: Optional[Union[str, Tuple[float, float, float]]] = None
        Location of sample in slide, defined either by descriptive string
        or xyz-coordinates.
    uid_generator: Callable[..., Uid] = generate_uid
        Function that can generate uids.

    Returns
    ----------
    Dataset
        Dataset containing a sample description.
    """
    sample_preparation_steps: List[SpecimenPreparationStep] = []
    sample_preparation_step = create_sample_preparation_step(
        sample_id,
        stainings,
        embedding_medium,
        fixative
    )
    sample_preparation_steps.append(sample_preparation_step)
    if specimen_id is not None:
        sample_sampling_step = create_sample_sampling_step(
            sample_id,
            specimen_id,
            specimen_sampling_method,
            specimen_type
        )
        sample_preparation_steps.append(sample_sampling_step)

    specimen = SpecimenDescription(
        specimen_id=sample_id,
        specimen_uid=uid_generator(),
        specimen_preparation_steps=sample_preparation_steps,
        specimen_location=location
    )
    if anatomical_sites is not None:
        specimen = add_anatomical_sites_to_specimen(
            specimen,
            anatomical_sites
        )

    return specimen


def create_sample_preparation_step(
    specimen_id: str,
    stainings: Sequence[str],
    embedding_medium: Optional[str],
    fixative: Optional[str]
) -> SpecimenPreparationStep:
    """Create SpecimenPreparationStep for a preparation step.

    Parameters
    ----------
    specimen_id: str
        ID of specimen that has been prepared.
    stainings: Sequence[str]
        Sequence of stainings used.
    embedding_medium: Optional[str] = None
        Embedding medium used.
    fixative: Optional[str] = None
        Fixative used.

    Returns
    ----------
    SpecimenPreparationStep
        SpecimenPreparationStep for a preparation step.
    """
    if embedding_medium is not None:
        embedding_medium_code = (
            SpecimenEmbeddingMediaCode(embedding_medium).code
        )
    else:
        embedding_medium_code = None
    if fixative is not None:
        fixative_code = SpecimenFixativesCode(fixative).code
    else:
        fixative_code = None

    processing_type = SpecimenPreparationProcedureCode('Staining').code
    processing_procedure = SpecimenStaining([
        SpecimenStainsCode(staining).code for staining in stainings
    ])
    sample_preparation_step = SpecimenPreparationStep(
        specimen_id=specimen_id,
        processing_type=processing_type,
        processing_procedure=processing_procedure,
        embedding_medium=embedding_medium_code,
        fixative=fixative_code
    )
    return sample_preparation_step


def create_sample_sampling_step(
    sample_id: str,
    specimen_id: str,
    specimen_sampling_method: Optional[str],
    specimen_type: Optional[str]
) -> SpecimenPreparationStep:
    """Create SpecimenPreparationStep for a sampling step.

    Parameters
    ----------
    sample_id: str
        ID of sample that has been created.
    specimen_id: str
        ID of specimen that was sampled.
    specimen_sampling_method: Optional[str]
        Method used to sample specimen.
    specimen_type: Optional[str]
        Type of specimen that was sampled

    Returns
    ----------
    SpecimenPreparationStep
        SpecimenPreparationStep for a sampling step.
    """
    if specimen_sampling_method is None:
        raise ValueError(
            'Specimen sampling method required if '
            'specimen id is defined'
        )
    if specimen_type is None:
        raise ValueError(
            'Specimen type required if '
            'specimen id is defined'
        )
    specimen_type_code = AnatomicPathologySpecimenTypesCode(specimen_type)
    sampling_method_code = SpecimenSamplingProcedureCode(
        specimen_sampling_method
    )
    sample_sampling_step = SpecimenPreparationStep(
        specimen_id=sample_id,
        processing_type=SpecimenPreparationProcedureCode(
            'Sampling of tissue specimen'
        ).code,
        processing_procedure=SpecimenSampling(
                method=sampling_method_code.code,
                parent_specimen_id=specimen_id,
                parent_specimen_type=specimen_type_code.code
        )
    )
    return sample_sampling_step


def add_anatomical_sites_to_specimen(
    specimen: Dataset,
    anatomical_sites: Sequence[Tuple[Code, Sequence[Code]]]
) -> Dataset:
    """Add anatomical site and anatomical site modifier codes to specimen.

    Parameters
    ----------
    specimen: Dataset
        Dataset containing a specimen description
    anatomical_sites: Sequence[Tuple[Code, Sequence[Code]]]
        List of original anatomical sites, each defined by a code and
        list of modifier codes (can be empty).

    Returns
    ----------
    Dataset
        Dataset containing a specimen description.
    """
    anatomical_site_datasets: List[Dataset] = []
    for (anatomical_site, modifiers) in anatomical_sites:
        anatomical_site_dataset = (
            ConceptCode.from_code(anatomical_site).to_ds()
        )

        if modifiers != []:
            modifier_datasets: List[Dataset] = []
            for modifier in modifiers:
                modifier_dataset = Dataset()
                modifier_dataset.CodeValue = modifier.value
                modifier_dataset.CodingSchemeDesignator = (
                    modifier.scheme_designator
                )
                modifier_dataset.CodeMeaning = modifier.meaning
                modifier_datasets.append(modifier_dataset)

            (
                anatomical_site_dataset
                .PrimaryAnatomicStructureModifierSequence
            ) = DicomSequence(modifier_datasets)
        anatomical_site_datasets.append(anatomical_site_dataset)
    specimen.PrimaryAnatomicStructureSequence = DicomSequence(
        anatomical_site_datasets
    )
    return specimen


def create_specimen_module(
    slide_id: str,
    samples: Union[Dataset, Sequence[Dataset]]
) -> Dataset:
    """Create specimen module.

    Parameters
    ----------
    slide_id: str.
        Identifier for the slide.
    samples: Union[Dataset, Sequence[Dataset]]
        Single or list of sample descriptions (should be created with
        create_sample())

    Returns
    ----------
    Dataset
        Dataset containing specimen module.
    """
    dataset = Dataset()
    dataset.ContainerIdentifier = slide_id

    container_type_code_sequence = Dataset()
    container_type_code_sequence.CodeValue = '258661006'
    container_type_code_sequence.CodingSchemeDesignator = 'SCT'
    container_type_code_sequence.CodeMeaning = 'Slide'
    dataset.ContainerTypeCodeSequence = (
        DicomSequence([container_type_code_sequence])
    )

    container_component_sequence = Dataset()
    container_component_sequence.ContainerComponentMaterial = 'GLASS'
    container_component_type_code_sequence = Dataset()
    container_component_type_code_sequence.CodeValue = '433472003'
    container_component_type_code_sequence.CodingSchemeDesignator = 'SCT'
    container_component_type_code_sequence.CodeMeaning = (
        'Microscope slide coverslip'
    )
    container_component_sequence.ContainerComponentTypeCodeSequence = (
        DicomSequence([container_component_type_code_sequence])
    )
    dataset.ContainerComponentSequence = (
        DicomSequence([container_component_sequence])
    )
    if not isinstance(samples, Sequence):
        samples = [samples]
    specimen_description_sequence = (
        DicomSequence(samples)
    )
    dataset.SpecimenDescriptionSequence = specimen_description_sequence

    return dataset


def create_brightfield_optical_path_module(
    magnification: Optional[float] = None
) -> Dataset:
    """Create optical path module for brightfield illumination conditions.

    Parameters
    ----------
    magnification: Optional[float] = None
        Objective magnification.

    Returns
    ----------
    Dataset
        Dataset containing optical path module for brightfield illumination.
    """
    dataset = Dataset()
    optical_path_item = Dataset()
    optical_path_item.OpticalPathIdentifier = '0'
    illumination_type_code_sequence = Dataset()
    illumination_type_code_sequence.CodeValue = '111744'
    illumination_type_code_sequence.CodingSchemeDesignator = 'DCM'
    illumination_type_code_sequence.CodeMeaning = (
        'Brightfield illumination'
    )
    optical_path_item.IlluminationTypeCodeSequence = DicomSequence(
        [illumination_type_code_sequence]
    )
    illumination_color_code_sequence = Dataset()
    illumination_color_code_sequence.CodeValue = 'R-102C0'
    illumination_color_code_sequence.CodingSchemeDesignator = 'SRT'
    illumination_color_code_sequence.CodeMeaning = 'Full Spectrum'
    optical_path_item.IlluminationColorCodeSequence = DicomSequence(
        [illumination_color_code_sequence]
    )
    if magnification is not None:
        optical_path_item.ObjectiveLensPower = magnification

    dataset.OpticalPathSequence = DicomSequence([optical_path_item])

    return dataset


def create_default_modules(
    uid_generator: Callable[..., Uid] = generate_uid
) -> List[Dataset]:
    """Return default module dataset for testing.

    Parameters
    ----------
    uid_generator: Callable[..., Uid]
        Function that can generate Uids.

    Returns
    ----------
    List[Dataset]
        Default module datasets.
    """
    modules: List[Dataset] = []

    # Generic study module
    modules.append(create_study_module())

    # Generic patient module
    modules.append(create_patient_module())

    # Generic device module
    modules.append(create_device_module())

    # Generic specimen module
    modules.append(create_specimen_module(
        'Unkown',
        samples=[create_sample(
            sample_id='Unkown',
            stainings=['water soluble eosin stain', 'hematoxylin stain'],
            uid_generator=uid_generator
        )]
    ))

    # Generic optical path sequence
    modules.append(create_brightfield_optical_path_module())

    return modules



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math
import os
from abc import ABCMeta
from ctypes import c_uint32
from pathlib import Path
from typing import List, Optional, Sequence, Union

import numpy as np
from PIL import Image
from pydicom import Dataset
from pydicom.uid import UID as Uid
from wsidicom import (WsiDicom, WsiDicomLabels, WsiDicomLevels,
                      WsiDicomOverviews)
from wsidicom.geometry import Point, Size, SizeMm, Region
from wsidicom.wsidicom import WsiDicom
from wsidicom.errors import WsiDicomNotFoundError

from wsidicomizer.common import MetaDicomizer, MetaImageData
from wsidicomizer.dataset import create_base_dataset
from wsidicomizer.encoding import Encoder, create_encoder

if os.name == 'nt':  # On windows, add path to openslide to dll path
    try:
        openslide_dir = os.environ['OPENSLIDE']
    except KeyError:
        raise ValueError(
            "Enviroment variable 'OPENSLIDE'"
            "needs to be set to OpenSlide bin path"
        )
    try:
        os.add_dll_directory(openslide_dir)
    except AttributeError:
        os.environ['PATH'] = (
            openslide_dir + os.pathsep + os.environ['PATH']
        )

from openslide import OpenSlide
from openslide._convert import argb2rgba as convert_argb_to_rgba
from openslide.lowlevel import (ArgumentError, _read_associated_image,
                                _read_region, get_associated_image_dimensions,
                                get_associated_image_names)

"""
OpenSlideWrapper uses private functions from OpenSlide to get image data as
numpy arrays instead of pillow images. The private functions
(_read_associated_image, _read_region) are used to get raw data from the
OpenSlide C API. We consider this safe, as these directly map to the Openslide
C API and are thus not likely  to change.
"""


class OpenSlideImageData(MetaImageData, metaclass=ABCMeta):
    def __init__(
        self,
        open_slide: OpenSlide,
        encoder: Encoder
    ):
        """Wraps a OpenSlide image to ImageData.

        Parameters
        ----------
        open_slide: OpenSlide
            OpenSlide object to wrap.
        encoded: Encoder
            Encoder to use.
        """
        super().__init__(encoder)

        self._open_slide = open_slide

    @property
    def files(self) -> List[Path]:
        return [Path(self._open_slide._filename)]

    @property
    def transfer_syntax(self) -> Uid:
        """The uid of the transfer syntax of the image."""
        return self._encoder.transfer_syntax

    @staticmethod
    def _make_transparent_pixels_white(image_data: np.ndarray) -> np.ndarray:
        """Return image data where all pixels with transparency is replaced
        by white pixels. Openslide returns fully transparent pixels with
        RGBA-value 0, 0, 0, 0 for 'sparse' areas. At the edge to 'sparse' areas
        there can also be partial transparency. This function 'aggresively'
        removes all transparent pixels (instead of calculating RGB-values
        with transparency for partial transparent pixels) as it is much,
        simpler, faster, and the partial transparency is at the edge of the
        ROIs.

        Parameters
        ----------
        image_data: np.ndarray
             Image data in RGBA pixel format to remove transparency from.

        Returns
        ----------
        image_data: np.ndarray
            Image data in RGBA pixel format without transparency.
        """
        transparency = image_data[:, :, 3]
        image_data[transparency != 255, :] = 255
        return image_data

    def close(self) -> None:
        """Close the open slide object, if not already closed."""
        try:
            self._open_slide.close()
        except ArgumentError:
            # Slide already closed
            pass


class OpenSlideAssociatedImageData(OpenSlideImageData):
    def __init__(
        self,
        open_slide: OpenSlide,
        image_type: str,
        encoder: Encoder
    ):
        """Wraps a OpenSlide associated image (label or overview) to ImageData.

        Parameters
        ----------
        open_slide: OpenSlide
            OpenSlide object to wrap.
        image_type: str
            Type of image to wrap.
        encoded: Encoder
            Encoder to use.
        """
        super().__init__(open_slide, encoder)
        self._image_type = image_type
        if image_type not in get_associated_image_names(self._open_slide._osr):
            raise ValueError(f"{image_type} not in {self._open_slide}")

        width, height = get_associated_image_dimensions(
            self._open_slide._osr,
            image_type
        )
        buffer = (width * height * c_uint32)()
        _read_associated_image(self._open_slide._osr, image_type, buffer)
        image_data: np.ndarray = np.frombuffer(buffer, dtype=np.uint8)
        image_data.shape = (width, height, 4)
        image_data = self._make_transparent_pixels_white(image_data)
        self._encoded_image = self._encode(image_data)
        convert_argb_to_rgba(image_data)
        self._decoded_image = Image.fromarray(image_data).convert('RGB')
        (height, width) = image_data.shape[0:2]
        self._image_size = Size(width, height)

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self.image_size

    @property
    def pixel_spacing(self) -> SizeMm:
        """Size of the pixels in mm/pixel."""
        # TODO figure out pixel spacing for label and overview in openslide.
        return SizeMm(1, 1)

    @property
    def pyramid_index(self) -> int:
        """The pyramidal index in relation to the base layer."""
        return 0

    def _get_encoded_tile(
        self,
        tile: Point,
        z: float,
        path: str
    ) -> bytes:
        if tile != Point(0, 0):
            raise ValueError
        return self._encoded_image

    def _get_decoded_tile(
        self,
        tile: Point,
        z: float,
        path: str
    ) -> Image.Image:
        if tile != Point(0, 0):
            raise ValueError
        return self._decoded_image


class OpenSlideLevelImageData(OpenSlideImageData):
    def __init__(
        self,
        open_slide: OpenSlide,
        level_index: int,
        tile_size: int,
        encoder: Encoder
    ):
        super().__init__(open_slide, encoder)
        """Wraps a OpenSlide level to ImageData.

        Parameters
        ----------
        open_slide: OpenSlide
            OpenSlide object to wrap.
        level_index: int
            Level in OpenSlide object to wrap
        tile_size: int
            Output tile size.
        encoded: Encoder
            Encoder to use.
        """
        self._tile_size = Size(tile_size, tile_size)
        self._open_slide = open_slide
        self._level_index = level_index
        self._image_size = Size.from_tuple(
            self._open_slide.level_dimensions[self._level_index]
        )
        self._downsample = int(
            self._open_slide.level_downsamples[self._level_index]
        )
        self._pyramid_index = int(math.log2(self.downsample))

        base_mpp_x = float(self._open_slide.properties['openslide.mpp-x'])
        base_mpp_y = float(self._open_slide.properties['openslide.mpp-y'])
        self._pixel_spacing = SizeMm(
            base_mpp_x * self.downsample / 1000.0,
            base_mpp_y * self.downsample / 1000.0
        )

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self._tile_size

    @property
    def pixel_spacing(self) -> SizeMm:
        """Size of the pixels in mm/pixel."""
        return self._pixel_spacing

    @property
    def downsample(self) -> int:
        """Downsample facator for level."""
        return self._downsample

    @property
    def pyramid_index(self) -> int:
        """The pyramidal index in relation to the base layer."""
        return self._pyramid_index

    def stitch_tiles(
        self,
        region: Region,
        path: str,
        z: float
    ) -> Image.Image:
        """Overrides ImageData stitch_tiles() to read reagion directly from
        openslide object.

        Parameters
        ----------
        region: Region
             Pixel region to stitch to image
        path: str
            Optical path
        z: float
            Z coordinate

        Returns
        ----------
        Image.Image
            Stitched image
        """
        if path not in self.optical_paths:
            raise WsiDicomNotFoundError(f"Optical path {path}", str(self))
        if z not in self.focal_planes:
            raise WsiDicomNotFoundError(f"Z {z}", str(self))
        if region.size.width < 0 or region.size.height < 0:
            raise ValueError('Negative size not allowed')

        location_in_base_level = region.start * self.downsample

        buffer = (region.size.width * region.size.height * c_uint32)()
        _read_region(
            self._open_slide._osr,
            buffer,
            location_in_base_level.x,
            location_in_base_level.y,
            self._level_index,
            region.size.width,
            region.size.height
        )
        tile_data: np.ndarray = np.frombuffer(buffer, dtype=np.uint8)
        tile_data.shape = (region.size.width, region.size.height, 4)
        tile_data = self._make_transparent_pixels_white(tile_data)
        convert_argb_to_rgba(tile_data)
        return Image.fromarray(tile_data).convert('RGB')

    def _get_tile(self, tile: Point, flip: bool = False) -> np.ndarray:
        """Return tile as np array. Transparency is removed. Optionally the
        pixel format can be flipped to RGBA, suitable for opening with PIL.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        flip: bool
            If to flip the pixel format from ARGB to RGBA.

        Returns
        ----------
        np.ndarray
            Numpy array of tile.
        """
        tile_point_in_base_level = tile * self.downsample * self._tile_size
        if self._tile_size.width < 0 or self._tile_size.height < 0:
            raise ValueError('Negative size not allowed')
        buffer = (self._tile_size.width * self._tile_size.height * c_uint32)()
        _read_region(
            self._open_slide._osr,
            buffer,
            tile_point_in_base_level.x,
            tile_point_in_base_level.y,
            self._level_index,
            self._tile_size.width,
            self._tile_size.height
        )
        tile_data: np.ndarray = np.frombuffer(buffer, dtype=np.uint8)
        tile_data.shape = (self._tile_size.width, self._tile_size.height, 4)
        tile_data = self._make_transparent_pixels_white(tile_data)
        if flip:
            convert_argb_to_rgba(tile_data)
        return tile_data

    def _get_encoded_tile(
        self,
        tile: Point,
        z: float,
        path: str
    ) -> bytes:
        """Return image bytes for tile. Transparency is removed and tile is
        encoded as jpeg.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        bytes
            Tile bytes.
        """
        if z not in self.focal_planes or path not in self.optical_paths:
            raise ValueError
        return self._encode(self._get_tile(tile))

    def _get_decoded_tile(
        self,
        tile_point: Point,
        z: float,
        path: str
    ) -> Image.Image:
        """Return Image for tile. Image mode is RGB.

        Parameters
        ----------
        tile_point: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        Image.Image
            Tile as Image.
        """
        if z not in self.focal_planes or path not in self.optical_paths:
            raise ValueError
        tile = self._get_tile(tile_point, True)
        return Image.fromarray(tile).convert('RGB')


class OpenSlideDicomizer(MetaDicomizer):
    @classmethod
    def open(
        cls,
        filepath: str,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        tile_size: Optional[int] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        encoding_format: str = 'jpeg',
        encoding_quality: int = 90,
        jpeg_subsampling: str = '422'
    ) -> WsiDicom:
        """Open openslide file in filepath as WsiDicom object. Note that
        created instances always has a random UID.

        Parameters
        ----------
        filepath: str
            Path to tiff file
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
            Module datasets to use in files. If none, use default modules.
        tile_size: Optional[int]
            Tile size to use if not defined by file.
        include_levels: Sequence[int] = None
            Levels to include. If None, include all levels.
        include_label: bool = True
            Inclube label.
        include_overview: bool = True
            Include overview.
        include_confidential: bool = True
            Include confidential metadata. Not implemented.
        encoding_format: str = 'jpeg'
            Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.
        encoding_quality: int = 90
            Quality to use if re-encoding. Do not use > 95 for jpeg. Use 100
            for lossless jpeg2000.
        jpeg_subsampling: str = '422'
            Subsampling option if using jpeg for re-encoding. Use '444' for
            no subsampling, '422' for 2x2 subsampling.

        Returns
        ----------
        WsiDicom
            WsiDicom object of openslide file in filepath.
        """
        if tile_size is None:
            raise ValueError("Tile size required for open slide")
        JCS_EXT_BGRA = 9
        encoder = create_encoder(
            encoding_format,
            encoding_quality,
            subsampling=jpeg_subsampling,
            colorspace=JCS_EXT_BGRA
        )
        base_dataset = create_base_dataset(modules)
        slide = OpenSlide(filepath)
        instance_number = 0
        level_instances = [
            cls._create_instance(
                OpenSlideLevelImageData(
                    slide,
                    level_index,
                    tile_size,
                    encoder
                ),
                base_dataset,
                'VOLUME',
                instance_number+level_index
            )
            for level_index in range(slide.level_count)
            if include_levels is None or level_index in include_levels
        ]
        instance_number += len(level_instances)
        if include_label and 'label' in slide.associated_images:
            label_instances = [cls._create_instance(
                OpenSlideAssociatedImageData(slide, 'label', encoder),
                base_dataset,
                'LABEL',
                instance_number
            )]
        else:
            label_instances = []
        instance_number += len(label_instances)
        if include_overview and 'macro' in slide.associated_images:
            overview_instances = [cls._create_instance(
                OpenSlideAssociatedImageData(slide, 'macro', encoder),
                base_dataset,
                'OVERVIEW',
                instance_number
            )]
        else:
            overview_instances = []
        levels = WsiDicomLevels.open(level_instances)
        labels = WsiDicomLabels.open(label_instances)
        overviews = WsiDicomOverviews.open(overview_instances)
        return cls(levels, labels, overviews)

    @staticmethod
    def is_supported(filepath: str) -> bool:
        """Return True if file in filepath is supported by OpenSlide."""
        return OpenSlide.detect_format(str(filepath)) is not None



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse
import json
import os
from pathlib import Path

from pydicom.dataset import Dataset

from wsidicomizer.interface import WsiDicomizer


def main():
    parser = argparse.ArgumentParser(
        description=('Convert combatible wsi file to DICOM')
    )

    parser.add_argument(
        '-i', '--input',
        type=Path,
        required=True,
        help='Path to input wsi file.'
    )
    parser.add_argument(
        '-o', '--output',
        type=Path,
        help=(
            'Path to output folder. Folder will be created and must not '
            'excist. If not specified a folder named after the input file is '
            'created in the same path.'
        )
    )
    parser.add_argument(
        '-t', '--tile-size',
        type=int,
        default=512,
        help=(
            'Tile size (same for width and height). Required for ndpi and '
            'openslide formats E.g. 512'
        )
    )
    parser.add_argument(
        '-d', '--dataset',
        type=Path,
        help=(
            'Path to json DICOM dataset. Can be used to define additional '
            'DICOM modules to include in the files'
        )
    )
    parser.add_argument(
        '-l', '--levels',
        type=int,
        nargs='+',
        help=(
            'Pyramid levels to include, if not all. E.g. 0 1 for base and '
            'first pyramid layer. '
        ),
    )
    parser.add_argument(
        '--no-label',
        action="store_true",
        help='If not to include label'
    )
    parser.add_argument(
        '--no-overview',
        action="store_true",
        help='If not to include overview'
    )
    parser.add_argument(
        '--no-confidential',
        action="store_true",
        help='If not to include confidential metadata'
    )
    parser.add_argument(
        '-w', '--workers',
        type=int,
        default=os.cpu_count(),
        help='Number of worker threads to use'
    )
    parser.add_argument(
        '--chunk-size',
        type=int,
        default=100,
        help='Number of tiles to give each worker at a time'
    )
    parser.add_argument(
        '--format',
        type=str,
        default='jpeg',
        help="Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'."
    )
    parser.add_argument(
        '--quality',
        type=int,
        default=90,
        help=(
            "Quality to use if re-encoding. Do not use > 95 for jpeg. "
            "Use 100 for lossless jpeg2000."
        )
    )
    parser.add_argument(
        '--subsampling',
        type=str,
        default='422',
        help=(
            "Subsampling option if using jpeg for re-encoding. Use '444' for "
            "no subsampling, '422' for 2x2 subsampling."
        )
    )

    args = parser.parse_args()
    if not args.dataset:
        dataset = None
    else:
        json_file = open(args.dataset)
        dataset = Dataset.from_json(json.load(json_file))
    if not args.levels:
        levels = None
    else:
        levels = args.levels

    WsiDicomizer.convert(
        filepath=args.input,
        output_path=args.output,
        modules=dataset,
        tile_size=args.tile_size,
        include_levels=levels,
        include_label=not args.no_label,
        include_overview=not args.no_overview,
        include_confidential=not args.no_confidential,
        workers=args.workers,
        chunk_size=args.chunk_size,
        encoding_format=args.format,
        encoding_quality=args.quality,
        jpeg_subsampling=args.subsampling
    )


if __name__ == "__main__":
    main()



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Union, Optional

import numpy as np
from imagecodecs import jpeg2k_encode, jpeg8_encode
from pydicom.uid import UID, JPEGBaseline8Bit, JPEG2000Lossless, JPEG2000


class Encoder(metaclass=ABCMeta):
    @property
    @abstractmethod
    def transfer_syntax(self) -> UID:
        raise NotImplementedError

    @property
    @abstractmethod
    def quality(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def encode(self, data: np.ndarray) -> bytes:
        raise NotImplementedError


class JpegEncoder(Encoder):
    def __init__(
        self,
        quality: int = 90,
        subsampling: Optional[str] = '422',
        colorspace: Optional[Union[int, str]] = None
    ) -> None:
        self._quality = quality
        self._subsampling = subsampling
        self._colorspace = colorspace
        self._outcolorspace = 'YCBCR'

    @property
    def transfer_syntax(self) -> UID:
        return JPEGBaseline8Bit

    @property
    def quality(self) -> int:
        return self._quality

    def encode(self, data: np.ndarray) -> bytes:
        if data.dtype != np.dtype(np.uint8):
            data = (data * 255 / np.iinfo(data.dtype).max).astype(np.uint8)
        return jpeg8_encode(
            data,
            level=self._quality,
            colorspace=self._colorspace,
            outcolorspace=self._outcolorspace,
            subsampling=self._subsampling
        )


class Jpeg2000Encoder(Encoder):
    def __init__(
        self,
        quality: int = 90,
    ) -> None:
        self._quality = quality
        if self.quality == 100:
            self._transfer_syntax = JPEG2000Lossless
        else:
            self._transfer_syntax = JPEG2000

    @property
    def transfer_syntax(self) -> UID:
        return self._transfer_syntax

    @property
    def quality(self) -> int:
        return self._quality

    def encode(self, data: np.ndarray) -> bytes:
        return jpeg2k_encode(
            data,
            level=self._quality
        )


def create_encoder(
    format: str,
    quality: int,
    subsampling: Optional[str] = None,
    colorspace: Optional[Union[int, str]] = None
) -> Encoder:
    if format == 'jpeg':
        return JpegEncoder(
            quality=quality,
            subsampling=subsampling,
            colorspace=colorspace
        )
    elif format == 'jpeg2000':
        return Jpeg2000Encoder(
            quality=quality
        )
    raise ValueError("Encoder format must be 'jpeg' or 'jpeg2000'")



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from xml.etree import ElementTree
from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from czifile import CziFile, DirectoryEntryDV
from PIL import Image
from pydicom import Dataset
from pydicom.uid import UID as Uid
from wsidicom import (WsiDicom, WsiDicomLabels, WsiDicomLevels,
                      WsiDicomOverviews)
from wsidicom.geometry import Point, Region, Size, SizeMm
from wsidicom.wsidicom import WsiDicom

from wsidicomizer.common import MetaImageData, MetaDicomizer
from wsidicomizer.dataset import create_base_dataset
from wsidicomizer.encoding import Encoder, create_encoder


def get_element(element: ElementTree.Element, tag: str) -> ElementTree.Element:
    found_element = element.find(tag)
    if found_element is None:
        raise ValueError(f"Tag {tag} not found in element")
    return found_element


def get_nested_element(
    element: ElementTree.Element,
    tags: List[str]
) -> ElementTree.Element:
    found_element = element
    for tag in tags:
        found_element = found_element.find(tag)
        if found_element is None:
            raise ValueError(f"Tag {tag} not found in element")
    return found_element


def get_text_from_element(
    element: ElementTree.Element,
    tag: str,
    default: Optional[str] = None
) -> str:
    try:
        element = get_element(element, tag)
        text = element.text
        if text is None:
            raise ValueError("Text not found in element")
    except ValueError:
        if default is not None:
            return default
        raise ValueError(f"Tag {tag} or text not found in element")
    return text


class CziImageData(MetaImageData):
    _default_z = 0

    def __init__(
        self,
        filepath: str,
        tile_size: int,
        encoder: Encoder
    ) -> None:
        """Wraps a czi file to ImageData. Multiple z, c, or pyramid levels are
        currently not supported.

        Parameters
        ----------
        filepath: str
            Path to czi file to wrap.
        tile_size: int
            Output tile size.
        encoded: Encoder
            Encoded to use.
        """
        self._filepath = filepath
        self._czi = CziFile(filepath)
        self._czi._fh.lock = True
        self._samples_per_pixel = self._get_samples_per_pixel()

        self._dtype = self._czi.dtype
        super().__init__(encoder)
        self._tile_size = Size(tile_size, tile_size)
        self._image_size = Size(self._czi.shape[4], self._czi.shape[3])
        self._image_origin = Point(self._czi.start[4], self._czi.start[3])
        self._focal_plane_mapping = self._get_focal_plane_mapping()
        self._channel_mapping = self._get_channel_mapping()
        self._block_directory = self._czi.filtered_subblock_directory
        self._tile_directory = self._create_tile_directory()
        self._blank_decoded_tile = Image.fromarray(self._create_blank_tile())
        self._blank_encoded_tile = self._encode(self._create_blank_tile())
        self._pixel_spacing = self._get_pixel_spacing()
        self._focal_planes = sorted(self._focal_plane_mapping)

    @property
    def pyramid_index(self) -> int:
        return 0

    @property
    def files(self) -> List[Path]:
        return [Path(self._filepath)]

    @property
    def transfer_syntax(self) -> Uid:
        return self._encoder.transfer_syntax

    @property
    def pixel_spacing(self) -> SizeMm:
        return self._pixel_spacing

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self._tile_size

    @property
    def tiled_size(self) -> Size:
        return self.image_size / self.tile_size

    @property
    def focal_planes(self) -> List[float]:
        """Focal planes avaiable in the image defined in um."""
        return self._focal_planes

    @property
    def optical_paths(self) -> List[str]:
        """Optical paths avaiable in the image."""
        return self._channel_mapping

    @property
    def blank_decoded_tile(self) -> Image.Image:
        return self._blank_decoded_tile

    @property
    def blank_encoded_tile(self) -> bytes:
        return self._blank_encoded_tile

    @property
    def image_origin(self) -> Point:
        """Return coordinate of the top-left of the image."""
        return self._image_origin

    @property
    def block_directory(self) -> List[DirectoryEntryDV]:
        """Return list of DirectoryEntryDV sorted by mosaic index."""
        return self._block_directory

    @property
    def tile_directory(self) -> Dict[Tuple[Point, float, str], Sequence[int]]:
        """Return dict of tiles with mosaic indices."""
        return self._tile_directory

    @property
    def metadata(self) -> str:
        metadata_xml = self._czi.metadata()
        if metadata_xml is None or not isinstance(metadata_xml, str):
            raise ValueError("No metadata string in file")
        return metadata_xml

    @property
    def samples_per_pixel(self) -> int:
        return self._samples_per_pixel

    @staticmethod
    def detect_format(filepath: Path) -> Optional[str]:
        try:
            czi = CziFile(filepath)
            czi.close()
            return 'czi'
        except ValueError:
            return None

    def _get_scaling(
        self
    ) -> Tuple[Optional[float], Optional[float], Optional[float]]:
        scaling_elements = get_nested_element(
            ElementTree.fromstring(self.metadata),
            ['Metadata', 'Scaling', 'Items']
        )
        x: Optional[float] = None
        y: Optional[float] = None
        z: Optional[float] = None
        for distance in scaling_elements.findall('Distance'):
            dimension = distance.get('Id')
            # Value is in m per pixel, result in mm per pixel
            value = (
                float(get_text_from_element(distance, 'Value')) * pow(10, 6)
            )
            if dimension == 'X':
                x = value
            elif dimension == 'Y':
                y = value
            elif dimension == 'Z':
                z = value
        return x, y, z

    def _get_pixel_spacing(self) -> SizeMm:
        """Get pixel spacing (mm per pixel) from metadata"""
        x, y, _ = self._get_scaling()
        if x is None or y is None:
            raise ValueError("Could not find pixel spacing in metadata")
        return SizeMm(x, y)/1000

    def _create_blank_tile(self, fill: float = 1) -> np.ndarray:
        """Return blank tile in numpy array.

        Parameters
        ----------
        fill: float = 1
            Value to fill tile with.

        Returns
        ----------
        np.ndarray
            A blank tile as numpy array.
        """
        return np.full(
            self._size_to_numpy_shape(self.tile_size),
            fill * np.iinfo(self._czi.dtype).max,
            dtype=np.dtype(self._czi.dtype)
        )

    def _get_tile(
        self,
        tile_point: Point,
        z: float,
        path: str
    ) -> np.ndarray:
        """Return tile data as numpy array for tile.

        Parameters
        ----------
        tile_point: Point
            Tile coordinate to get.

        Returns
        ----------
        np.ndarray
            Tile as numpy array.
        """
        # A blank tile to paste blocks into
        image_data = self._create_blank_tile()
        if (tile_point, z, path) not in self.tile_directory:
            # Should not happen (get_decoded_tile() and get_enoded_tile()
            # should already have checked).
            return image_data

        for block_index in self.tile_directory[tile_point, z, path]:
            # For each block covering the tile
            block = self._block_directory[block_index]
            block_data: np.ndarray = block.data_segment().data()

            # Start and end coordiantes for block and tile
            block_start, block_size = self._get_block_start_and_size(block)
            block_end = block_start + block_size
            tile_start = tile_point * self.tile_size
            tile_end = (tile_point + 1) * self.tile_size

            # The block and tile both cover the region between these points
            tile_block_min_intersection = Point.max(tile_start, block_start)
            tile_block_max_intersection = Point.min(tile_end, block_end)

            # The intersects in relation to block and tile origin
            block_start_in_tile = tile_block_min_intersection - tile_start
            block_end_in_tile = tile_block_max_intersection - tile_start
            tile_start_in_block = tile_block_min_intersection - block_start
            tile_end_in_block = tile_block_max_intersection - block_start

            # Reshape the block data to remove leading 1-indices.
            block_data.shape = self._size_to_numpy_shape(block_size)
            # Paste in block data into tile.
            image_data[
                block_start_in_tile.y:block_end_in_tile.y,
                block_start_in_tile.x:block_end_in_tile.x,
                :
            ] = block_data[
                tile_start_in_block.y:tile_end_in_block.y,
                tile_start_in_block.x:tile_end_in_block.x,
                :
            ]
        return image_data

    def _get_decoded_tile(
        self,
        tile: Point,
        z: float,
        path: str
    ) -> Image.Image:
        """Return Image for tile.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        Image.Image
            Tile as Image.
        """
        if (tile, z, path) not in self.tile_directory:
            return self.blank_decoded_tile
        return Image.fromarray(self._get_tile(tile, z, path))

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        """Return image bytes for tile. Tile is encoded as jpeg.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        bytes
            Tile bytes.
        """
        if (tile, z, path) not in self.tile_directory:
            return self.blank_encoded_tile
        frame = self._get_tile(tile, z, path)
        if self._dtype != np.dtype(np.uint8):
            frame = (frame * 255 / np.iinfo(self._dtype)).astype(np.uint8)
        return self._encode(self._get_tile(tile, z, path))

    def close(self) -> None:
        """Close wrapped file."""
        self._czi.close()

    def _create_tile_directory(
        self
    ) -> Dict[Tuple[Point, float, str], Sequence[int]]:
        """Create a directory mapping tile points to list of block indices that
        cover the tile. This could be extended to also index z and c.

        Returns
        ----------
        Dict[Tuple[Point, int, int], Sequence[int]]:
            Directory of tile point, focal plane and channel as key and
            lists of block indices as item.
        """
        tile_directory: DefaultDict[Tuple[Point, float, str], List[int]] = (
            defaultdict(list)
        )

        for block_index, block in enumerate(self.block_directory):
            block_start, block_size, z, c = self._get_block_dimensions(block)
            tile_region = Region.from_points(
                block_start // self.tile_size,
                (block_start+block_size) // self.tile_size
            )
            for tile in tile_region.iterate_all(include_end=True):
                tile_directory[tile, z, c].append(block_index)

        return dict(tile_directory)

    def _get_block_start_and_size(
        self,
        block: DirectoryEntryDV
    ) -> Tuple[Point, Size]:
        """Return start coordinate and size for block realtive to image
        origin.

        Parameters
        ----------
        block: DirectoryEntryDV
            Block to get start and size from.

        Returns
        ----------
        Tuple[Point, Size]
            Start point corrdinate and size for block.
        """
        x_start: Optional[int] = None
        x_size: Optional[int] = None
        y_start: Optional[int] = None
        y_size: Optional[int] = None

        for dimension_entry in block.dimension_entries:
            if dimension_entry.dimension == 'X':
                x_start = dimension_entry.start
                x_size = dimension_entry.size
            elif dimension_entry.dimension == 'Y':
                y_start = dimension_entry.start
                y_size = dimension_entry.size
        if (
            x_start is None or x_size is None or
            y_start is None or y_size is None
        ):
            raise ValueError(
                "Could not determine position of block"
            )

        return (
            Point(x_start, y_start) - self.image_origin,
            Size(x_size, y_size)
        )

    def _get_block_dimensions(
        self,
        block: DirectoryEntryDV
    ) -> Tuple[Point, Size, float, str]:
        """Return start coordinate and size for block realtive to image
        origin. This could be extended to also get z and c.

        Parameters
        ----------
        block: DirectoryEntryDV
            Block to get start and size from.

        Returns
        ----------
        Tuple[Point, Size]
            Start point corrdinate and size for block.
        """
        start, size = self._get_block_start_and_size(block)
        z = 0.0
        c = '0'
        for dimension_entry in block.dimension_entries:
            if dimension_entry.dimension == 'X':
                x_start = dimension_entry.start
                x_size = dimension_entry.size
            elif dimension_entry.dimension == 'Y':
                y_start = dimension_entry.start
                y_size = dimension_entry.size
            elif dimension_entry.dimension == 'Z':
                z = self._focal_plane_mapping[dimension_entry.start]
            elif dimension_entry.dimension == 'C':
                c = self._channel_mapping[dimension_entry.start]
        if start is None or size is None or z is None or c is None:
            raise ValueError(
                "Could not determine position of block"
            )

        return start, size, z, c

    def _size_to_numpy_shape(self, size: Size) -> Tuple[int, int, int]:
        """Return a tuple for use with numpy.shape."""
        return size.height, size.width, self.samples_per_pixel

    def _get_focal_plane_mapping(self) -> List[float]:
        image = get_nested_element(
            ElementTree.fromstring(self.metadata),
            ["Metadata", 'Information', 'Image']
        )
        try:
            size_z = int(get_text_from_element(image, 'SizeZ', '0'))
            z_interval = get_nested_element(
                image,
                ['Dimensions', 'Z', 'Positions', 'Interval']
            )
            start = int(get_text_from_element(z_interval, 'Start'))
            increment = int(get_text_from_element(z_interval, 'Increment'))
            _, _, z_scale = self._get_scaling()
            if z_scale is None:
                raise ValueError("No z scale in metadata")
            start_z = start * z_scale
            step_z = (start+increment*size_z)*z_scale
            end_z = increment*z_scale
            return list(np.arange(start_z, end_z, step_z))
        except ValueError:
            return [0.0]

    def _get_channel_mapping(self) -> List[str]:
        channels = get_nested_element(
            ElementTree.fromstring(self.metadata),
            ['Metadata', 'Information', 'Image', 'Dimensions', 'Channels']
        )
        return [
            get_text_from_element(channel, 'Fluor')
            for channel in channels
        ]

    def _get_samples_per_pixel(self) -> int:
        return self._czi.shape[-1]


class CziDicomizer(MetaDicomizer):
    @classmethod
    def open(
        cls,
        filepath: str,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        tile_size: Optional[int] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        encoding_format: str = 'jpeg',
        encoding_quality: int = 90,
        jpeg_subsampling: str = '422'
    ) -> WsiDicom:
        """Open czi file in filepath as WsiDicom object. Note that created
        instances always has a random UID.

        Parameters
        ----------
        filepath: str
            Path to tiff file
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
            Module datasets to use in files. If none, use default modules.
        tile_size: Optional[int]
            Tile size to use if not defined by file.
        include_levels: Sequence[int] = None
            Levels to include. Not implemented.
        include_label: bool = True
            Inclube label. Not implemented.
        include_overview: bool = True
            Include overview. Not implemented.
        include_confidential: bool = True
            Include confidential metadata. Not implemented.
        encoding_format: str = 'jpeg'
            Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.
        encoding_quality: int = 90
            Quality to use if re-encoding. Do not use > 95 for jpeg. Use 100
            for lossless jpeg2000.
        jpeg_subsampling: str = '422'
            Subsampling option if using jpeg for re-encoding. Use '444' for
            no subsampling, '422' for 2x2 subsampling.

        Returns
        ----------
        WsiDicom
            WsiDicom object of czi file in filepath.
        """
        if tile_size is None:
            raise ValueError("Tile size required for open slide")
        encoder = create_encoder(
            encoding_format,
            encoding_quality,
            jpeg_subsampling
        )
        base_dataset = create_base_dataset(modules)
        base_level_instance = cls._create_instance(
            CziImageData(filepath, tile_size, encoder),
            base_dataset,
            'VOLUME',
            0
        )
        levels = WsiDicomLevels.open([base_level_instance])
        labels = WsiDicomLabels.open([])
        overviews = WsiDicomOverviews.open([])
        return cls(levels, labels, overviews)

    @staticmethod
    def is_supported(filepath: str) -> bool:
        """Return True if file in filepath is supported by CziFile."""
        return CziImageData.detect_format(Path(filepath)) is not None



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import Optional, Sequence, Union

import numpy as np
from pydicom import Dataset, config
from pydicom.dataset import Dataset
from pydicom.sequence import Sequence as DicomSequence
from pydicom.uid import UID, JPEGBaseline8Bit, generate_uid
from pydicom.valuerep import DSfloat
from wsidicom import ImageData, WsiDicom, WsiInstance
from wsidicom.instance import WsiDataset

from wsidicomizer.encoding import Encoder

from .dataset import get_image_type

config.enforce_valid_values = True
config.future_behavior()


class MetaImageData(ImageData, metaclass=ABCMeta):
    _default_z = 0

    def __init__(
        self,
        encoder: Encoder
    ):
        """Wraps a OpenTilePage to ImageData.

        Parameters
        ----------
        tiled_page: OpenTilePage
            OpenTilePage to wrap.
        encoded: Encoder
            Encoder to use.
        """
        self._encoder = encoder

    @property
    @abstractmethod
    def pyramid_index(self) -> int:
        """Should return pyramid level for image data."""
        raise NotImplementedError

    @property
    def samples_per_pixel(self) -> int:
        return 3

    @property
    def photometric_interpretation(self) -> str:
        # Should be derived from the used subsample format
        return 'YBR_FULL'

    def create_instance_dataset(
        self,
        base_dataset: Dataset,
        image_flavor: str,
        instance_number: int,
        transfer_syntax: UID,
        photometric_interpretation: str
    ) -> WsiDataset:
        """Return instance dataset for image_data based on base dataset.

        Parameters
        ----------
        base_dataset: Dataset
            Dataset common for all instances.
        image_flavor:
            Type of instance ('VOLUME', 'LABEL', 'OVERVIEW)
        instance_number: int
        transfer_syntax: UID
        photometric_interpretation: str

        Returns
        ----------
        WsiDataset
            Dataset for instance.
        """
        dataset = deepcopy(base_dataset)
        dataset.ImageType = get_image_type(
            image_flavor,
            self.pyramid_index
        )
        dataset.SOPInstanceUID = generate_uid(prefix=None)

        shared_functional_group_sequence = Dataset()
        pixel_measure_sequence = Dataset()
        pixel_measure_sequence.PixelSpacing = [
            DSfloat(self.pixel_spacing.width, True),
            DSfloat(self.pixel_spacing.height, True)
        ]
        pixel_measure_sequence.SpacingBetweenSlices = 0.0
        pixel_measure_sequence.SliceThickness = 0.0
        shared_functional_group_sequence.PixelMeasuresSequence = (
            DicomSequence([pixel_measure_sequence])
        )
        dataset.SharedFunctionalGroupsSequence = DicomSequence(
            [shared_functional_group_sequence]
        )
        dataset.DimensionOrganizationType = 'TILED_FULL'
        dataset.TotalPixelMatrixColumns = self.image_size.width
        dataset.TotalPixelMatrixRows = self.image_size.height
        dataset.Columns = self.tile_size.width
        dataset.Rows = self.tile_size.height
        dataset.NumberOfFrames = (
            self.tiled_size.width
            * self.tiled_size.height
        )
        dataset.ImagedVolumeWidth = (
            self.image_size.width * self.pixel_spacing.width
        )
        dataset.ImagedVolumeHeight = (
            self.image_size.height * self.pixel_spacing.height
        )
        dataset.ImagedVolumeDepth = 0.0

        if transfer_syntax == JPEGBaseline8Bit:
            dataset.BitsAllocated = 8
            dataset.BitsStored = 8
            dataset.HighBit = 7
            dataset.PixelRepresentation = 0
            # dataset.LossyImageCompressionRatio = 1
            dataset.LossyImageCompressionMethod = 'ISO_10918_1'
        if photometric_interpretation == 'YBR_FULL':
            dataset.PhotometricInterpretation = photometric_interpretation
            dataset.SamplesPerPixel = 3

        dataset.PlanarConfiguration = 0

        dataset.InstanceNumber = instance_number
        dataset.FocusMethod = 'AUTO'
        dataset.ExtendedDepthOfField = 'NO'
        return WsiDataset(dataset)

    def _encode(self, image_data: np.ndarray) -> bytes:
        """Return image data encoded in jpeg using set quality and subsample
        options.

        Parameters
        ----------
        image_data: np.ndarray
            Image data to encode, in BGRA-pixel format.

        Returns
        ----------
        bytes
            Jpeg bytes.
        """
        return self._encoder.encode(image_data)


class MetaDicomizer(WsiDicom, metaclass=ABCMeta):
    """Metaclass for Dicomizers. Subclasses should implement is_supported() and
    open().
    """
    @staticmethod
    @abstractmethod
    def is_supported(path: str) -> bool:
        """Return True if file in filepath is supported by Dicomizer."""
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def open(
        cls,
        filepath: str,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        tile_size: Optional[int] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        encoding_format: str = 'jpeg',
        encoding_quality: int = 90,
        jpeg_subsampling: str = '422'
    ) -> WsiDicom:
        """Open file in filepath as WsiDicom object. Note that created
        instances always has a random UID.

        Parameters
        ----------
        filepath: str
            Path to tiff file
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
            Module datasets to use in files. If none, use default modules.
        tile_size: Optional[int]
            Tile size to use if not defined by file.
        include_levels: Sequence[int] = None
            Levels to include. If None, include all levels.
        include_label: bool = True
            Inclube label.
        include_overview: bool = True
            Include overview.
        include_confidential: bool = True
            Include confidential metadata.
        encoding_format: str = 'jpeg'
            Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.
        encoding_quality: int = 90
            Quality to use if re-encoding. Do not use > 95 for jpeg. Use 100
            for lossless jpeg2000.
        jpeg_subsampling: str = '422'
            Subsampling option if using jpeg for re-encoding. Use '444' for
            no subsampling, '422' for 2x2 subsampling.

        Returns
        ----------
        WsiDicom
            WsiDicom object of file in filepath.
        """
        raise NotImplementedError()

    @staticmethod
    def _create_instance(
        image_data: MetaImageData,
        base_dataset: Dataset,
        image_type: str,
        instance_number: int
    ) -> WsiInstance:
        """Create WsiInstance from MetaImageData.

        Parameters
        ----------
        image_data: ImageData
            Image data and metadata.
        base_dataset: Dataset
            Base dataset to include.
        image_type: str
            Type of instance to create.
        instance_number: int
            The number of the instance (in a series).

        Returns
        ----------
        WsiInstance
            Created WsiInstance.
        """
        instance_dataset = image_data.create_instance_dataset(
            base_dataset,
            image_type,
            instance_number,
            image_data.transfer_syntax,
            image_data.photometric_interpretation
        )

        return WsiInstance(
            instance_dataset,
            image_data
        )



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os
from pathlib import Path
from typing import Callable, List, Optional, Sequence, Type, Union

from pydicom.dataset import Dataset
from pydicom.uid import UID, generate_uid
from wsidicom import WsiDicom

from wsidicomizer.common import MetaDicomizer
from wsidicomizer.czi import CziDicomizer
from wsidicomizer.openslide import OpenSlideDicomizer
from wsidicomizer.opentile import OpenTileDicomizer

# List of supported Dicomizers in prioritization order.
SUPPORTED_TILE_SOURCES: List[Type[MetaDicomizer]] = [
    OpenTileDicomizer,
    CziDicomizer,
    OpenSlideDicomizer
]


class WsiDicomizer:
    """Interface for Dicomizing files."""
    @staticmethod
    def open(
        filepath: str,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        tile_size: Optional[int] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        encoding_format: str = 'jpeg',
        encoding_quality: int = 90,
        jpeg_subsampling: str = '422'
    ) -> WsiDicom:
        """Open data in file in filepath as WsiDicom.

        Parameters
        ----------
        filepath: str
            Path to file
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
            Module datasets to use in files. If none, use default modules.
        tile_size: int
            Tile size to use if not defined by file.
        include_levels: Sequence[int]
            Optional list of levels to include. Include all levels if None.
        include_label: bool
            Include label(s), default true.
        include_overwiew: bool
            Include overview(s), default true.
        include_confidential: bool = True
            Include confidential metadata.
        encoding_format: str = 'jpeg'
            Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.
        encoding_quality: int = 90
            Quality to use if re-encoding. Do not use > 95 for jpeg. Use 100
            for lossless jpeg2000.
        jpeg_subsampling: str = '422'
            Subsampling option if using jpeg for re-encoding. Use '444' for
            no subsampling, '422' for 2x2 subsampling.

        Returns
        ----------
        WsiDicom
            WsiDicom object of file.
        """
        selected_tile_source = next(
            (
                tile_source for tile_source in SUPPORTED_TILE_SOURCES
                if tile_source.is_supported(filepath)
            ),
            None
        )
        if selected_tile_source is None:
            raise NotImplementedError(f"{filepath} is not supported")

        return selected_tile_source.open(
            filepath,
            modules,
            tile_size,
            include_levels,
            include_label,
            include_overview,
            include_confidential,
            encoding_format,
            encoding_quality,
            jpeg_subsampling
        )

    @classmethod
    def convert(
        cls,
        filepath: str,
        output_path: Optional[str] = None,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        tile_size: Optional[int] = None,
        uid_generator: Callable[..., UID] = generate_uid,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        workers: Optional[int] = None,
        chunk_size: Optional[int] = None,
        encoding_format: str = 'jpeg',
        encoding_quality: int = 90,
        jpeg_subsampling: str = '422'
    ) -> List[str]:
        """Convert data in file to DICOM files in output path. Created
        instances get UID from uid_generator. Closes when finished.

        Parameters
        ----------
        filepath: str
            Path to file
        output_path: str = None
            Folder path to save files to.
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
            Module datasets to use in files. If none, use default modules.
        tile_size: int
            Tile size to use if not defined by file.
        uid_generator: Callable[..., UID] = generate_uid
             Function that can gernerate unique identifiers.
        include_levels: Sequence[int]
            Optional list of levels to include. Include all levels if None.
        include_label: bool
            Include label(s), default true.
        include_overwiew: bool
            Include overview(s), default true.
        include_confidential: bool = True
            Include confidential metadata.
        workers: Optional[int] = None,
            Maximum number of thread workers to use.
        chunk_size: Optional[int] = None,
            Chunk size (number of tiles) to process at a time. Actual chunk
            size also depends on minimun_chunk_size from image_data.
        encoding_format: str = 'jpeg'
            Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.
        encoding_quality: int = 90
            Quality to use if re-encoding. Do not use > 95 for jpeg. Use 100
            for lossless jpeg2000.
        jpeg_subsampling: str = '422'
            Subsampling option if using jpeg for re-encoding. Use '444' for
            no subsampling, '422' for 2x2 subsampling.

        Returns
        ----------
        List[str]
            List of paths of created files.
        """
        with cls.open(
            filepath,
            modules,
            tile_size,
            include_levels,
            include_label,
            include_overview,
            include_confidential,
            encoding_format,
            encoding_quality,
            jpeg_subsampling
        ) as wsi:
            if output_path is None:
                output_path = str(Path(filepath).parents[0].joinpath(
                    Path(filepath).stem
                ))
            try:
                os.mkdir(output_path)
            except FileExistsError:
                ValueError(f'Output path {output_path} already excists')
            created_files = wsi.save(
                output_path,
                uid_generator,
                workers,
                chunk_size
            )

        return [str(filepath) for filepath in created_files]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wsidicomizer']

package_data = \
{'': ['*']}

install_requires = \
['czifile>=2019.7.2,<2020.0.0',
 'highdicom>=0.9.2,<0.10.0',
 'imagecodecs>=2021.8.26,<2022.0.0',
 'numpy>=1.21.2,<2.0.0',
 'openslide-python>=1.1.2,<2.0.0',
 'opentile>=0.1.0,<0.2.0',
 'pydicom>=2.2.1,<3.0.0',
 'wsidicom>=0.1.0,<0.2.0']

entry_points = \
{'console_scripts': ['wsidicomizer = wsidicomizer.cli:main']}

setup_kwargs = {
    'name': 'wsidicomizer',
    'version': '0.1.0',
    'description': 'Tool for converting wsi-files to DICOM',
    'long_description': "# *wsidicomizer*\n*wsidicomizer* is a Python library for converting files wsi files to DICOM. The aims of the project are:\n- Provide lossless conversion for files supported by opentile.\n- Provide 'as good as possible' conversion for other formats.\n- Simplify the encoding of wsi metadata into DICOM.\n\n## Installation\n***Download wsidicomizer from git***\n```console\n$ pip install wsidicomizer\n```\n\n***Install OpenSlide***\nInstructions for how to install OpenSlide is avaiable on https://openslide.org/download/\nFor Windows, you need also need add OpenSlide's bin-folder to the environment variable 'OPENSLIDE'\n\n***Install libjpeg-turbo***\nInstall libjpeg-turbo either as binary from https://libjpeg-turbo.org/ or using your package manager.\nFor Windows, you also need to add libjpeg-turbo's bin-folder to the environment variable 'TURBOJPEG'\n\n## Important note\nPlease note that this is an early release and the API is not frozen yet. Function names and functionality is prone to change.\n\n## Requirements\n*wsidicomizer* requires python >=3.7 and uses numpy, pydicom, highdicom, imagecodecs, openslide-python, PyTurboJPEG, opentile, and wsidicom.\n\n## Limitations\nFiles with z-stacks or multiple focal paths are currently not supported. DICOM properties related to slice thickness, focal plane thickness, and imaged volume are saved as 0 and not with proper values.\n\n## Basic cli-usage\n***Convert a wsi-file into DICOM using cli-interface***\n```console\nwsidicomizer -i 'path_to_wsi_file' -o 'path_to_output_folder'\n```\n### Arguments:\n~~~~\n-i, --input, path to input wsi file\n-o, --output, path to output folder\n-t, --tile-size, required depending on input format\n-d, --dataset, optional path to json file defining base dataset\n-l, --levels, optional levels to include\n-w, --workers, number of threads to use\n--chunk-size, number of tiles to give each worker at a time\n--format, encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'\n--quality, quality to use if re-encoding.\n--subsampling, subsampling option to use if re-encoding.\n~~~~\n### Flags\n~~~~\n--no-label, do not include label(s)\n--no-overview, do not include overview(s)\n--no-confidential, do not include confidential metadata from image\n~~~~\nUsing the no-confidential-flag properties according to [DICOM Basic Confidentiality Profile](https://dicom.nema.org/medical/dicom/current/output/html/part15.html#table_E.1-1) are not included in the output file. Properties otherwise included are currently:\n* Acquisition DateTime\n* Device Serial Number\n\n## Basic notebook-usage\n***Create module datasets (Optional)***\n```python\nfrom wsidicomizer.dataset import create_device_module, create_sample, create_specimen_module, create_brightfield_optical_path_module, create_patient_module, create_study_module\ndevice_module = create_device_module(\n    manufacturer='Scanner manufacturer',\n    model_name='Scanner model name',\n    serial_number='Scanner serial number',\n    software_versions=['Scanner software versions']\n)\nsample = create_sample(\n    sample_id='sample id',\n    embedding_medium='Paraffin wax',\n    fixative='Formalin',\n    stainings=['hematoxylin stain', 'water soluble eosin stain']\n)\nspecimen_module = create_specimen_module(\n    slide_id='slide id',\n    samples=[sample]\n)\noptical_module = create_brightfield_optical_path_module()\npatient_module = create_patient_module()\nstudy_module = create_study_module()\n\n```\n\n***Convert a wsi-file into DICOM using python-interface***\n```python\nfrom wsidicomizer import WsiDicomizer\ncreated_files = WsiDicomizer.convert(\n    path_to_wsi_file,\n    path_to_output_folder,\n    [device_module, specimen_module, optical_module, patient_module, study_module],\n    tile_size\n)\n```\ntile_size is required for Ndpi- and OpenSlide-files.\n\n***Import a wsi file as a WsiDicom object.***\n```python\nfrom wsidicomizer import WsiDicomizer\nwsi = WsiDicomizer.import_tiff(path_to_wsi_file)\nregion = wsi.read_region((1000, 1000), 6, (200, 200))\nwsi.close()\n```\n\n## Other DICOM python tools\n- [pydicom](https://pydicom.github.io/)\n- [highdicom](https://github.com/MGHComputationalPathology/highdicom)\n- [wsidicom](https://github.com/imi-bigpicture/wsidicom)\n\n## Contributing\nWe welcome any contributions to help improve this tool for the WSI DICOM community!\n\nWe recommend first creating an issue before creating potential contributions to check that the contribution is in line with the goals of the project. To submit your contribution, please issue a pull request on the imi-bigpicture/wsidicomizer repository with your changes for review.\n\nOur aim is to provide constructive and positive code reviews for all submissions. The project relies on gradual typing and roughly follows PEP8. However, we are not dogmatic. Most important is that the code is easy to read and understand.\n\n## TODOs\n* Packaging of libjpeg-turbo into an 'ready-to-use' distribution.\n* Look into if OpenSlide python will provide a 'ready-to-use' distribution.\n* Interface for coding annotations (geometrical, diagnosis using for example structured reporting).\n\n## Acknowledgement\n*wsidicomizer*: Copyright 2021 Sectra AB, licensed under Apache 2.0.\n\nThis project is part of a project that has received funding from the Innovative Medicines Initiative 2 Joint Undertaking under grant agreement No 945358. This Joint Undertaking receives support from the European Union’s Horizon 2020 research and innovation programme and EFPIA. IMI website: www.imi.europa.eu",
    'author': 'Erik O Gabrielsson',
    'author_email': 'erik.o.gabrielsson@sectra.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/imi-bigpicture/wsidicomizer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Union

from PIL import Image
from pydicom import Dataset
from pydicom.uid import JPEG2000, UID, JPEGBaseline8Bit
from wsidicom import (WsiDicom, WsiDicomLabels, WsiDicomLevels,
                      WsiDicomOverviews, WsiInstance)
from wsidicom.geometry import Point, Size, SizeMm

from opentile import OpenTile
from opentile.common import OpenTilePage, Tiler
from wsidicomizer.common import MetaDicomizer, MetaImageData
from wsidicomizer.dataset import create_base_dataset, populate_base_dataset
from wsidicomizer.encoding import Encoder, create_encoder


class OpenTileImageData(MetaImageData):
    def __init__(
        self,
        tiled_page: OpenTilePage,
        encoder: Encoder
    ):
        """Wraps a OpenTilePage to ImageData.

        Parameters
        ----------
        tiled_page: OpenTilePage
            OpenTilePage to wrap.
        encoded: Encoder
            Encoder to use.
        """
        super().__init__(encoder)
        self._tiled_page = tiled_page

        self._needs_transcoding = not self.is_supported_transfer_syntax()
        if self.needs_transcoding:
            self._transfer_syntax = self._encoder.transfer_syntax
        else:
            self._transfer_syntax = self.get_transfer_syntax()
        self._image_size = Size(*self._tiled_page.image_size.to_tuple())
        self._tile_size = Size(*self._tiled_page.tile_size.to_tuple())
        self._tiled_size = Size(*self._tiled_page.tiled_size.to_tuple())
        self._pixel_spacing = SizeMm(
            *self._tiled_page.pixel_spacing.to_tuple()
        )

    def __str__(self) -> str:
        return f"{type(self).__name__} for page {self._tiled_page}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._tiled_page})"

    @property
    def files(self) -> List[Path]:
        return [Path(self._tiled_page.filepath)]

    @property
    def transfer_syntax(self) -> UID:
        """The uid of the transfer syntax of the image."""
        return self._transfer_syntax

    @property
    def needs_transcoding(self) -> bool:
        """Return true if image data requires transcoding for Dicom
        compatibilty."""
        return self._needs_transcoding

    @property
    def native_compression(self) -> str:
        """Return compression method used in image data."""
        return self._tiled_page.compression

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self._tile_size

    @property
    def pixel_spacing(self) -> SizeMm:
        """Size of the pixels in mm/pixel."""
        return self._pixel_spacing

    @property
    def focal_planes(self) -> List[float]:
        """Focal planes avaiable in the image defined in um."""
        return [self._tiled_page.focal_plane]

    @property
    def optical_paths(self) -> List[str]:
        """Optical paths avaiable in the image."""
        return [self._tiled_page.optical_path]

    @property
    def suggested_minimum_chunk_size(self) -> int:
        """Return suggested minumum chunk size for optimal performance with
        get_encoeded_tiles()."""
        return self._tiled_page.suggested_minimum_chunk_size

    @property
    def pyramid_index(self) -> int:
        """The pyramidal index in relation to the base layer."""
        return self._tiled_page.pyramid_index

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        """Return image bytes for tile. Returns transcoded tile if
        non-supported encoding.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        bytes
            Tile bytes.
        """
        if z not in self.focal_planes or path not in self.optical_paths:
            raise ValueError
        if self.needs_transcoding:
            decoded_tile = self._tiled_page.get_decoded_tile(tile.to_tuple())
            return self._encode(decoded_tile)
        return self._tiled_page.get_tile(tile.to_tuple())

    def _get_decoded_tile(
        self,
        tile: Point,
        z: float,
        path: str
    ) -> Image.Image:
        """Return Image for tile.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        Image.Image
            Tile as Image.
        """
        if z not in self.focal_planes or path not in self.optical_paths:
            raise ValueError
        return Image.fromarray(
            self._tiled_page.get_decoded_tile(tile.to_tuple())
        )

    def get_encoded_tiles(
        self,
        tiles: Sequence[Point],
        z: float,
        path: str
    ) -> List[bytes]:
        """Return list of image bytes for tiles. Returns transcoded tiles if
        non-supported encoding.

        Parameters
        ----------
        tiles: Sequence[Point]
            Tile positions to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        Iterator[List[bytes]]
            Iterator of tile bytes.
        """
        if z not in self.focal_planes or path not in self.optical_paths:
            raise ValueError
        tiles_tuples = [tile.to_tuple() for tile in tiles]
        if not self.needs_transcoding:
            return self._tiled_page.get_tiles(tiles_tuples)
        decoded_tiles = self._tiled_page.get_decoded_tiles(tiles_tuples)
        return [self._encode(tile) for tile in decoded_tiles]

    def close(self) -> None:
        self._tiled_page.close()

    def is_supported_transfer_syntax(self) -> bool:
        """Return true if image data is encoded with Dicom-supported transfer
        syntax."""
        try:
            self.get_transfer_syntax()
            return True
        except NotImplementedError:
            return False

    def get_transfer_syntax(self) -> UID:
        """Return transfer syntax (Uid) for compression type in image data."""
        compression = self.native_compression
        if compression == 'COMPRESSION.JPEG':
            return JPEGBaseline8Bit
        elif compression == 'COMPRESSION.APERIO_JP2000_RGB':
            return JPEG2000
        raise NotImplementedError(
            f'Not supported compression {compression}'
        )


class OpenTileDicomizer(MetaDicomizer):
    @classmethod
    def open(
        cls,
        filepath: str,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        tile_size: Optional[int] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        encoding_format: str = 'jpeg',
        encoding_quality: int = 90,
        jpeg_subsampling: str = '422'
    ) -> WsiDicom:
        """Open tiff file in filepath as WsiDicom object. Note that created
        instances always has a random UID.

        Parameters
        ----------
        filepath: str
            Path to tiff file
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
            Module datasets to use in files. If none, use default modules.
        tile_size: Optional[int]
            Tile size to use if not defined by file.
        include_levels: Sequence[int] = None
            Levels to include. If None, include all levels.
        include_label: bool = True
            Inclube label.
        include_overview: bool = True
            Include overview.
        include_confidential: bool = True
            Include confidential metadata.
        encoding_format: str = 'jpeg'
            Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.
        encoding_quality: int = 90
            Quality to use if re-encoding. Do not use > 95 for jpeg. Use 100
            for lossless jpeg2000.
        jpeg_subsampling: str = '422'
            Subsampling option if using jpeg for re-encoding. Use '444' for
            no subsampling, '422' for 2x2 subsampling.

        Returns
        ----------
        WsiDicom
            WsiDicom object of tiff file in filepath.
        """
        encoder = create_encoder(
            encoding_format,
            encoding_quality,
            subsampling=jpeg_subsampling
        )
        base_dataset = create_base_dataset(modules)
        tiler = OpenTile.open(filepath, tile_size)
        level_instances, label_instances, overview_instances = cls._open_tiler(
            tiler,
            encoder,
            base_dataset,
            include_levels=include_levels,
            include_label=include_label,
            include_overview=include_overview,
            include_confidential=include_confidential
        )
        levels = WsiDicomLevels.open(level_instances)
        labels = WsiDicomLabels.open(label_instances)
        overviews = WsiDicomOverviews.open(overview_instances)
        return cls(levels, labels, overviews)

    @staticmethod
    def is_supported(filepath: str) -> bool:
        """Return True if file in filepath is supported by OpenTile."""
        return OpenTile.detect_format(Path(filepath)) is not None

    @classmethod
    def _open_tiler(
        cls,
        tiler: Tiler,
        encoder: Encoder,
        base_dataset: Dataset,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True
    ) -> Tuple[List[WsiInstance], List[WsiInstance], List[WsiInstance]]:
        """Open tiler to produce WsiInstances.

        Parameters
        ----------
        tiler: Tiler
            Tiler that can produce WsiInstances.
        encoder: Encoder
            Encoder to use for re-encoding.
        base_dataset: Dataset
            Base dataset to include in files.
        include_levels: Optional[Sequence[int]] = None
            Optional list of levels to include. Include all levels if None.
        include_label: bool = True
            Include label(s), default true.
        include_overwiew: bool = True
            Include overview(s), default true.
        include_confidential: bool = True
            Include confidential metadata.

        Returns
        ----------
        Tuple[List[WsiInstance], List[WsiInstance], List[WsiInstance]]
            Lists of created level, label and overivew instances.
        """
        base_dataset = populate_base_dataset(
            tiler,
            base_dataset,
            include_confidential
        )
        instance_number = 0
        level_instances = [
            cls._create_instance(
                OpenTileImageData(level, encoder),
                base_dataset,
                'VOLUME',
                instance_number+index
            )
            for index, level in enumerate(tiler.levels)
            if include_levels is None or level.pyramid_index in include_levels
        ]
        instance_number += len(level_instances)
        label_instances = [
            cls._create_instance(
                OpenTileImageData(label, encoder),
                base_dataset,
                'LABEL',
                instance_number+index
            )
            for index, label in enumerate(tiler.labels)
            if include_label
        ]
        instance_number += len(level_instances)
        overview_instances = [
            cls._create_instance(
                OpenTileImageData(overview, encoder),
                base_dataset,
                'OVERVIEW',
                instance_number+index
            )
            for index, overview in enumerate(tiler.overviews)
            if include_overview
        ]

        return level_instances, label_instances, overview_instances
